@@ -1,4 +1,4 @@
-# Copyright (c) 2006-2017  Andrey Golovigin
+# Copyright (c) 2006-2018  Andrey Golovigin
 #
 # Permission is hereby granted, free of charge, to any person obtaining
 # a copy of this software and associated documentation files (the
@@ -19,13 +19,20 @@
 # TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 # SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 
+from __future__ import absolute_import, unicode_literals
+
 import re
 
+import six
+from pybtex.bibtex.exceptions import BibTeXError
+from pybtex.py3compat import fix_unicode_literals_in_doctest
 from pybtex.utils import pairwise
+from pybtex import py3compat
 
 whitespace_re = re.compile(r'(\s)')
 purify_special_char_re = re.compile(r'^\\[A-Za-z]+')
 
+
 def wrap(string, width=79, subsequent_indent='  '):
     """
     Wrap long string into multiple lines by inserting line breaks.
@@ -89,8 +96,12 @@ def wrap(string, width=79, subsequent_indent='  '):
     return '\n'.join(line.rstrip() for line in iter_lines(string))
 
 
+@py3compat.python_2_unicode_compatible
 class BibTeXString(object):
-    def __init__(self, chars, level=0):
+    def __init__(self, chars, level=0, max_level=100):
+        if level > max_level:
+            raise BibTeXError('too many nested braces')
+
         self.level = level
         self.is_closed = False
         self.contents = list(self.find_closing_brace(iter(chars)))
@@ -136,7 +147,7 @@ class BibTeXString(object):
         return ''.join(self.traverse(open=lambda string: '{', close=lambda string: '}'))
 
     def inner_string(self):
-        return ''.join(str(child) for child in self.contents)
+        return ''.join(six.text_type(child) for child in self.contents)
 
 
 def change_case(string, mode):
@@ -237,7 +248,7 @@ def bibtex_substring(string, start, length):
         end0 = len(string) + start + 1
         start0 = end0 - length
     else: # start == 0:
-        return ''
+        return u''
     return string[start0:end0]
 
 
@@ -411,87 +422,136 @@ def scan_bibtex_string(string):
     )
 
 
+@fix_unicode_literals_in_doctest
 def split_name_list(string):
     """
     Split a list of names, separated by ' and '.
 
     >>> split_name_list('Johnson and Peterson')
-    ['Johnson', 'Peterson']
+    [u'Johnson', u'Peterson']
     >>> split_name_list('Johnson AND Peterson')
-    ['Johnson', 'Peterson']
+    [u'Johnson', u'Peterson']
     >>> split_name_list('Johnson AnD Peterson')
-    ['Johnson', 'Peterson']
+    [u'Johnson', u'Peterson']
     >>> split_name_list('Armand and Peterson')
-    ['Armand', 'Peterson']
+    [u'Armand', u'Peterson']
     >>> split_name_list('Armand and anderssen')
-    ['Armand', 'anderssen']
+    [u'Armand', u'anderssen']
     >>> split_name_list('{Armand and Anderssen}')
-    ['{Armand and Anderssen}']
+    [u'{Armand and Anderssen}']
     >>> split_name_list('What a Strange{ }and Bizzare Name! and Peterson')
-    ['What a Strange{ }and Bizzare Name!', 'Peterson']
+    [u'What a Strange{ }and Bizzare Name!', u'Peterson']
     >>> split_name_list('What a Strange and{ }Bizzare Name! and Peterson')
-    ['What a Strange and{ }Bizzare Name!', 'Peterson']
+    [u'What a Strange and{ }Bizzare Name!', u'Peterson']
     """
     return split_tex_string(string, ' [Aa][Nn][Dd] ')
 
 
+@fix_unicode_literals_in_doctest
+def _find_closing_brace(string):
+    """
+    >>> _find_closing_brace('')
+    (u'', u'')
+    >>> _find_closing_brace('no braces')
+    (u'no braces', u'')
+    >>> _find_closing_brace('brace at the end}')
+    (u'brace at the end}', u'')
+    >>> _find_closing_brace('two closing braces}}')
+    (u'two closing braces}', u'}')
+    >>> _find_closing_brace('two closing} braces} and some text')
+    (u'two closing}', u' braces} and some text')
+    >>> _find_closing_brace('more {nested{}}{braces}} and the rest}')
+    (u'more {nested{}}{braces}}', u' and the rest}')
+    """
+    up_to_brace = []
+    brace_level = 1
+    while brace_level >= 1:
+        next_brace = BRACE_RE.search(string)
+        if not next_brace:
+            break
+
+        up_to_brace.append(string[:next_brace.end()])
+        string = string[next_brace.end():]
+
+        if next_brace.group() == '{':
+            brace_level += 1
+        elif next_brace.group() == '}':
+            brace_level -= 1
+        else:
+            raise ValueError(next_brace.group())
+
+    if not up_to_brace:
+        up_to_brace, string = [string], ''
+    return ''.join(up_to_brace), string
+
+
+# "\ " is a "control space" in TeX, i. e. "a space that is not to be ignored"
+#     -- The TeXbook, Chapter 3: Controlling TeX, p 8
+# ~ is a space character, according to BibTeX
+# \~ is not a space character
+BIBTEX_SPACE_RE = re.compile(r'(?:\\ |\s|(?<!\\)~)+')
+BRACE_RE = re.compile(r'{|}')
+
+
+@fix_unicode_literals_in_doctest
 def split_tex_string(string, sep=None, strip=True, filter_empty=False):
     r"""Split a string using the given separator (regexp).
 
     Everything at brace level > 0 is ignored.
-    Separators at the edges of the string are ignored.
 
     >>> split_tex_string('')
     []
     >>> split_tex_string('     ')
     []
-    >>> split_tex_string('   ', ' ', strip=False, filter_empty=False)
-    [' ', ' ']
     >>> split_tex_string('.a.b.c.', r'\.')
-    ['.a', 'b', 'c.']
+    [u'', u'a', u'b', u'c', u'']
     >>> split_tex_string('.a.b.c.{d.}.', r'\.')
-    ['.a', 'b', 'c', '{d.}.']
+    [u'', u'a', u'b', u'c', u'{d.}', u'']
     >>> split_tex_string('Matsui      Fuuka')
-    ['Matsui', 'Fuuka']
+    [u'Matsui', u'Fuuka']
     >>> split_tex_string('{Matsui      Fuuka}')
-    ['{Matsui      Fuuka}']
+    [u'{Matsui      Fuuka}']
     >>> split_tex_string(r'Matsui\ Fuuka')
-    ['Matsui', 'Fuuka']
+    [u'Matsui', u'Fuuka']
     >>> split_tex_string('{Matsui\ Fuuka}')
-    ['{Matsui\\ Fuuka}']
+    [u'{Matsui\\ Fuuka}']
     >>> split_tex_string('a')
-    ['a']
+    [u'a']
     >>> split_tex_string('on a')
-    ['on', 'a']
+    [u'on', u'a']
+    >>> split_tex_string(r'Qui\~{n}onero-Candela, J.')
+    [u'Qui\\~{n}onero-Candela,', u'J.']
     """
 
     if sep is None:
-        # "\ " is a "control space" in TeX,
-        # i. e. "a space that is not to be ignored"
-        # The TeXbook, Chapter 3: Controlling TeX, p 8
-        sep = r'(\\ |[\s~])+'
+        sep = BIBTEX_SPACE_RE
         filter_empty = True
 
-    sep_re = re.compile(sep)
-    brace_level = 0
-    name_start = 0
+    sep = re.compile(sep)
+
     result = []
-    string_len = len(string)
-    pos = 0
-    for pos, char in enumerate(string):
-        if char == '{':
-            brace_level += 1
-        elif char == '}':
-            brace_level -= 1
-        elif brace_level == 0 and pos > 0:
-            match = sep_re.match(string[pos:])
-            if match:
-                sep_len = len(match.group())
-                if pos + sep_len < string_len:
-                    result.append(string[name_start:pos])
-                    name_start = pos + sep_len
-    if name_start < string_len:
-        result.append(string[name_start:])
+    word_parts = []
+
+    while True:
+        head, brace, string = string.partition('{')
+
+        if head:
+            head_parts = sep.split(head)
+            for word in head_parts[:-1]:
+                result.append(''.join(word_parts + [word]))
+                word_parts = []
+            word_parts.append(head_parts[-1])
+
+        if brace:
+            word_parts.append(brace)
+            up_to_closing_brace, string = _find_closing_brace(string)
+            word_parts.append(up_to_closing_brace)
+        else:
+            break
+
+    if word_parts:
+        result.append(''.join(word_parts))
+
     if strip:
         result = [part.strip() for part in result]
     if filter_empty:
@@ -521,7 +581,7 @@ def bibtex_first_letter(string):
 
     for char in BibTeXString(string):
         if char.startswith('\\') and char != '\\':
-            return '{{{0}}}'.format(char)
+            return u'{{{0}}}'.format(char)
         elif char.isalpha():
             return char
     return ''
@@ -549,4 +609,3 @@ def bibtex_abbreviate(string, delimiter=None, separator='-'):
     if delimiter is None:
         delimiter = '.-'
     return delimiter.join(_bibtex_abbreviate())
-
