@@ -1,5 +1,5 @@
 # vim: fileencoding=utf-8
-# Copyright (c) 2006-2017  Andrey Golovigin
+# Copyright (c) 2006-2018  Andrey Golovigin
 #
 # Permission is hereby granted, free of charge, to any person obtaining
 # a copy of this software and associated documentation files (the
@@ -20,11 +20,15 @@
 # TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 # SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 
+from __future__ import unicode_literals
+from __future__ import print_function
 import re
 
 from collections import Mapping
 from pybtex.plugin import find_plugin
 
+import six
+
 from pybtex.exceptions import PybtexError
 from pybtex.utils import (
     deprecated,
@@ -33,6 +37,7 @@ from pybtex.utils import (
 from pybtex.richtext import Text
 from pybtex.bibtex.utils import split_tex_string, scan_bibtex_string
 from pybtex.errors import report_error
+from pybtex.py3compat import fix_unicode_literals_in_doctest, python_2_unicode_compatible
 
 
 class BibliographyDataError(PybtexError):
@@ -83,7 +88,7 @@ class BibliographyData(object):
             self.citations = CaseInsensitiveSet()
         if entries:
             if isinstance(entries, Mapping):
-                entries = iter(entries.items())
+                entries = entries.items()
             for (key, entry) in entries:
                 self.add_entry(key, entry)
         if preamble:
@@ -149,7 +154,6 @@ class BibliographyData(object):
         if key in self.entries:
             report_error(BibliographyDataError('repeated bibliograhpy entry: %s' % key))
             return
-        entry.collection = self
         entry.key = self.get_canonical_key(key)
         self.entries[entry.key] = entry
         try:
@@ -164,6 +168,7 @@ class BibliographyData(object):
         for key, entry in entries:
             self.add_entry(key, entry)
 
+    @fix_unicode_literals_in_doctest
     def _get_crossreferenced_citations(self, citations, min_crossrefs):
         """
         Get cititations not cited explicitly but referenced by other citations.
@@ -176,21 +181,21 @@ class BibliographyData(object):
         >>> list(data._get_crossreferenced_citations([], min_crossrefs=1))
         []
         >>> list(data._get_crossreferenced_citations(['main_article'], min_crossrefs=1))
-        ['xrefd_arcicle']
+        [u'xrefd_arcicle']
         >>> list(data._get_crossreferenced_citations(['Main_article'], min_crossrefs=1))
-        ['xrefd_arcicle']
+        [u'xrefd_arcicle']
         >>> list(data._get_crossreferenced_citations(['main_article'], min_crossrefs=2))
         []
         >>> list(data._get_crossreferenced_citations(['xrefd_arcicle'], min_crossrefs=1))
         []
 
-        >>> data2 = BibliographyData(data.entries, wanted_entries=list(data.entries.keys()))
+        >>> data2 = BibliographyData(data.entries, wanted_entries=data.entries.keys())
         >>> list(data2._get_crossreferenced_citations([], min_crossrefs=1))
         []
         >>> list(data2._get_crossreferenced_citations(['main_article'], min_crossrefs=1))
-        ['xrefd_arcicle']
+        [u'xrefd_arcicle']
         >>> list(data2._get_crossreferenced_citations(['Main_article'], min_crossrefs=1))
-        ['xrefd_arcicle']
+        [u'xrefd_arcicle']
         >>> list(data2._get_crossreferenced_citations(['main_article'], min_crossrefs=2))
         []
         >>> list(data2._get_crossreferenced_citations(['xrefd_arcicle'], min_crossrefs=1))
@@ -225,6 +230,7 @@ class BibliographyData(object):
                 citation_set.add(canonical_crossref)
                 yield canonical_crossref
 
+    @fix_unicode_literals_in_doctest
     def _expand_wildcard_citations(self, citations):
         """
         Expand wildcard citations (\citation{*} in .aux file).
@@ -239,15 +245,15 @@ class BibliographyData(object):
         >>> list(data._expand_wildcard_citations([]))
         []
         >>> list(data._expand_wildcard_citations(['*']))
-        ['uno', 'dos', 'tres', 'cuatro']
+        [u'uno', u'dos', u'tres', u'cuatro']
         >>> list(data._expand_wildcard_citations(['uno', '*']))
-        ['uno', 'dos', 'tres', 'cuatro']
+        [u'uno', u'dos', u'tres', u'cuatro']
         >>> list(data._expand_wildcard_citations(['dos', '*']))
-        ['dos', 'uno', 'tres', 'cuatro']
+        [u'dos', u'uno', u'tres', u'cuatro']
         >>> list(data._expand_wildcard_citations(['*', 'uno']))
-        ['uno', 'dos', 'tres', 'cuatro']
+        [u'uno', u'dos', u'tres', u'cuatro']
         >>> list(data._expand_wildcard_citations(['*', 'DOS']))
-        ['uno', 'dos', 'tres', 'cuatro']
+        [u'uno', u'dos', u'tres', u'cuatro']
 
         """
 
@@ -300,15 +306,16 @@ class BibliographyData(object):
 
         .. versionadded:: 0.19
         """
-        if isinstance(file, str):
+        if isinstance(file, six.string_types):
             filename = file
         else:
             filename = getattr(file, 'name', None)
         writer = find_plugin('pybtex.database.output', bib_format, filename=filename)(**kwargs)
         return writer.write_file(self, file)
 
+    @fix_unicode_literals_in_doctest
     def lower(self):
-        '''
+        u'''
         Return another :py:class:`.BibliographyData` with all identifiers converted to lowercase.
 
         >>> data = parse_string("""
@@ -322,18 +329,18 @@ class BibliographyData(object):
         ...     }
         ... """, 'bibtex')
         >>> data_lower = data.lower()
-        >>> list(data_lower.entries.keys())
-        ['obrazy', 'elegie']
-        >>> for entry in list(data_lower.entries.values()):
+        >>> data_lower.entries.keys()
+        [u'obrazy', u'elegie']
+        >>> for entry in data_lower.entries.values():
         ...     entry.key
-        ...     list(entry.persons.keys())
-        ...     list(entry.fields.keys())
-        'obrazy'
-        ['author']
-        ['title']
-        'elegie'
-        ['author']
-        ['title']
+        ...     entry.persons.keys()
+        ...     entry.fields.keys()
+        u'obrazy'
+        [u'author']
+        [u'title']
+        u'elegie'
+        [u'author']
+        [u'title']
 
         '''
 
@@ -346,27 +353,6 @@ class BibliographyData(object):
         )
 
 
-class FieldDict(OrderedCaseInsensitiveDict):
-    def __init__(self, parent, *args, **kwargw):
-        self.parent = parent
-        super(FieldDict, self).__init__(*args, **kwargw)
-
-    def __getitem__(self, key):
-        try:
-            return super(FieldDict, self).__getitem__(key)
-        except KeyError:
-            if key in self.parent.persons:
-                persons = self.parent.persons[key]
-                return ' and '.join(str(person) for person in persons)
-            elif 'crossref' in self:
-                return self.parent.get_crossref().fields[key]
-            else:
-                raise KeyError(key)
-
-    def lower(self):
-        return type(self)(self.parent, self.iteritems_lower())
-
-
 class RichFieldProxyDict(Mapping):
     def __init__(self, fields):
         self._fields = fields
@@ -397,9 +383,6 @@ class Entry(object):
     """A dictionary of entry fields.
     The dictionary is ordered and case-insensitive."""
 
-    rich_fields = None
-    """A dictionary of entry fields, converted to :ref:`rich text <rich-text>`."""
-
     persons = None
     """
     A dictionary of entry persons, by their roles.
@@ -407,27 +390,18 @@ class Entry(object):
     The most often used roles are ``'author'`` and ``'editor'``.
     """
 
-    collection = None
     """A reference to the containing :py:class:`.BibliographyData` object. Used to resolve crossrefs."""
 
-    def __init__(self, type_, fields=None, persons=None, collection=None):
+    def __init__(self, type_, fields=None, persons=None):
         if fields is None:
             fields = {}
         if persons is None:
             persons = {}
         self.type = type_.lower()
         self.original_type = type_
-
-        self.fields = FieldDict(self, fields)
-        self.rich_fields = RichFieldProxyDict(self.fields)
-
+        self.fields = OrderedCaseInsensitiveDict(fields)
         self.persons = OrderedCaseInsensitiveDict(persons)
 
-        self.collection = collection
-
-        # for BibTeX interpreter
-        self.vars = {}
-
     def __eq__(self, other):
         if not isinstance(other, Entry):
             return super(Entry, self) == other
@@ -438,9 +412,8 @@ class Entry(object):
         )
 
     def __repr__(self):
-        # representing fields as FieldDict causes problems with representing
-        # fields.parent, so represent it as a list of tuples
-        repr_fields = repr(list(self.fields.items()))
+        # represent the fields as a list of tuples for simplicity
+        repr_fields = repr(self.fields.items())
 
         return 'Entry({type_}, fields={fields}, persons={persons})'.format(
             type_=repr(self.type),
@@ -448,9 +421,6 @@ class Entry(object):
             persons=repr(self.persons),
         )
 
-    def get_crossref(self):
-        return self.collection.entries[self.fields['crossref']]
-
     def add_person(self, person, role):
         self.persons.setdefault(role, []).append(person)
 
@@ -459,20 +429,54 @@ class Entry(object):
             self.type,
             fields=self.fields.lower(),
             persons=self.persons.lower(),
-            collection=self.collection,
         )
 
+    def _find_person_field(self, role):
+        persons = self.persons[role]
+        return ' and '.join(six.text_type(person) for person in persons)
 
+    def _find_crossref_field(self, name, bib_data):
+        if bib_data is None or 'crossref' not in self.fields:
+            raise KeyError(name)
+        referenced_entry = bib_data.entries[self.fields['crossref']]
+        return referenced_entry._find_field(name, bib_data)
+
+    def _find_field(self, name, bib_data=None):
+        """
+        Find the field with the given ``name`` according to this rules:
+
+        - If the given field ``name`` in in ``self.fields``, just return
+          self.fields[name].
+
+        - Otherwise, if ``name`` is ``"authors"`` or ``"editors"`` (or any other
+          person role), return the list of names as a string, separated by
+          ``" and "``.
+
+        - Otherwise, if this entry has a ``crossreff`` field, look up for the
+          cross-referenced entry and try to find its field with the given
+          ``name``.
+        """
+        try:
+            return self.fields[name]
+        except KeyError:
+            try:
+                return self._find_person_field(name)
+            except KeyError:
+                return self._find_crossref_field(name, bib_data)
+
+
+@python_2_unicode_compatible
+@fix_unicode_literals_in_doctest
 class Person(object):
     """A person or some other person-like entity.
 
     >>> knuth = Person('Donald E. Knuth')
     >>> knuth.first_names
-    ['Donald']
+    [u'Donald']
     >>> knuth.middle_names
-    ['E.']
+    [u'E.']
     >>> knuth.last_names
-    ['Knuth']
+    [u'Knuth']
 
     """
 
@@ -552,6 +556,7 @@ class Person(object):
         self.lineage_names.extend(split_tex_string(lineage))
 
     @property
+    @fix_unicode_literals_in_doctest
     def bibtex_first_names(self):
         """A list of first and middle names together.
         (BibTeX treats all middle names as first.)
@@ -562,50 +567,51 @@ class Person(object):
 
         >>> knuth = Person('Donald E. Knuth')
         >>> knuth.bibtex_first_names
-        ['Donald', 'E.']
+        [u'Donald', u'E.']
         """
         return self.first_names + self.middle_names
 
+    @fix_unicode_literals_in_doctest
     def _parse_string(self, name):
         """Extract various parts of the name from a string.
 
         >>> p = Person('Avinash K. Dixit')
         >>> print(p.first_names)
-        ['Avinash']
+        [u'Avinash']
         >>> print(p.middle_names)
-        ['K.']
+        [u'K.']
         >>> print(p.prelast_names)
         []
         >>> print(p.last_names)
-        ['Dixit']
+        [u'Dixit']
         >>> print(p.lineage_names)
         []
-        >>> print(str(p))
+        >>> print(six.text_type(p))
         Dixit, Avinash K.
-        >>> p == Person(str(p))
+        >>> p == Person(six.text_type(p))
         True
         >>> p = Person('Dixit, Jr, Avinash K. ')
         >>> print(p.first_names)
-        ['Avinash']
+        [u'Avinash']
         >>> print(p.middle_names)
-        ['K.']
+        [u'K.']
         >>> print(p.prelast_names)
         []
         >>> print(p.last_names)
-        ['Dixit']
+        [u'Dixit']
         >>> print(p.lineage_names)
-        ['Jr']
-        >>> print(str(p))
+        [u'Jr']
+        >>> print(six.text_type(p))
         Dixit, Jr, Avinash K.
-        >>> p == Person(str(p))
+        >>> p == Person(six.text_type(p))
         True
 
         >>> p = Person('abc')
         >>> print(p.first_names, p.middle_names, p.prelast_names, p.last_names, p.lineage_names)
-        [] [] [] ['abc'] []
+        [] [] [] [u'abc'] []
         >>> p = Person('Viktorov, Michail~Markovitch')
         >>> print(p.first_names, p.middle_names, p.prelast_names, p.last_names, p.lineage_names)
-        ['Michail'] ['Markovitch'] [] ['Viktorov'] []
+        [u'Michail'] [u'Markovitch'] [] [u'Viktorov'] []
         """
         def process_first_middle(parts):
             try:
@@ -713,20 +719,21 @@ class Person(object):
         return ', '.join(part for part in (von_last, jr, first) if part)
 
     def __repr__(self):
-        return 'Person({0})'.format(repr(str(self)))
+        return 'Person({0})'.format(repr(six.text_type(self)))
 
     def get_part_as_text(self, type):
         names = getattr(self, type + '_names')
         return ' '.join(names)
 
+    @fix_unicode_literals_in_doctest
     def get_part(self, type, abbr=False):
         """Get a list of name parts by `type`.
 
         >>> knuth = Person('Donald E. Knuth')
         >>> knuth.get_part('first')
-        ['Donald']
+        [u'Donald']
         >>> knuth.get_part('last')
-        ['Knuth']
+        [u'Knuth']
         """
 
         names = getattr(self, type + '_names')
@@ -843,7 +850,7 @@ def parse_file(file, bib_format=None, **kwargs):
     .. versionadded:: 0.19
     """
 
-    if isinstance(file, str):
+    if isinstance(file, six.string_types):
         filename = file
     else:
         filename = getattr(file, 'name', None)
@@ -878,4 +885,3 @@ def parse_bytes(value, bib_format, **kwargs):
 
     parser = find_plugin('pybtex.database.input', bib_format)(**kwargs)
     return parser.parse_bytes(value)
-
